package com.vernaio.example;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.LineNumberReader;
import java.nio.charset.StandardCharsets;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Optional;

import org.supercsv.io.CsvListReader;
import org.supercsv.prefs.CsvPreference;

import com.fasterxml.jackson.core.JsonGenerationException;
import com.fasterxml.jackson.core.JsonParseException;
import com.fasterxml.jackson.databind.JsonMappingException;
import com.fasterxml.jackson.databind.ObjectMapper;
import com.fasterxml.jackson.datatype.jdk8.Jdk8Module;
import com.vernaio.aivis.engine.v2.base.sdk.dto.DtoFloatDataPoint;
import com.vernaio.aivis.engine.v2.base.sdk.exception.EngineException;
import com.vernaio.aivis.engine.v2.base.sdk.exception.MissingLibraryException;
import com.vernaio.aivis.engine.v2.base.sdk.exception.WrongFlavourException;
import com.vernaio.aivis.engine.v2.dependency.analysis.sdk.DependencyAnalysisData;

public class Utils {
	
	private static final ObjectMapper MAPPER = new ObjectMapper().registerModule(new Jdk8Module());
	
	public static class FloatSignal {
		
		private final String signalId;
		
		private final List<DtoFloatDataPoint> dataPoints;
		
		protected FloatSignal(final String signalId) {
			this.signalId = signalId;
			this.dataPoints = new ArrayList<DtoFloatDataPoint>();
		}
		
		public List<DtoFloatDataPoint> getDataPoints() {
			return this.dataPoints;
		}
		
		public String getSignalId() {
			return signalId;
		}
	}
	
	/*
	 * The following method knows very well how the data csv in this example looks like, reads it and at the end constructs the data context:
	 * Signals are added to the data context by naming their signal id and a list of their data points
	 * (DtoFloatDataPoint for float signals, DtoStringDataPoint for string signals, would be analogously for boolean signals).
	 * The method can be adapted if the data comes in a different format, e.g. via parquet files.
	 */
	public static DependencyAnalysisData readCsvConvertToAivisFormat(final File file) throws FileNotFoundException, IOException, WrongFlavourException,
		MissingLibraryException, EngineException {
		final HashMap<String, FloatSignal> floatSignals = new HashMap<String, FloatSignal>();
		final CsvPreference csvPreference = new CsvPreference.Builder('"', ',', "\n").ignoreEmptyLines(false).build();
		
		System.out.println(String.format("\t... reading file '%s' ...", file.getPath()));
		
		// get line numbers to show nice progress bar
		int lineCount = 0;
		try (LineNumberReader count = new LineNumberReader(new FileReader(file))) {
			count.skip(Long.MAX_VALUE);
			lineCount = count.getLineNumber();
		}
		
		// read file and construct lists of data points
		try (
			final InputStream stream = new FileInputStream(file);
			final InputStreamReader reader = new InputStreamReader(stream, StandardCharsets.UTF_8);
			final CsvListReader csvReader = new CsvListReader(reader, csvPreference)
		) {
			
			// the signals ids are in the csv's header
			final String[] header = csvReader.getHeader(true);
			for (int i = 1; i < header.length; i++) {
				final String signalId = header[i];
				floatSignals.put(signalId, new FloatSignal(signalId));
			}
			
			System.out.print("\t... processing signals ...");
			
			// show nice progress bar
			int progressBar = -1;
			int line = 0;
			
			// read values by index
			List<String> values;
			while ((values = csvReader.read()) != null) {
				
				// show nice progress bar
				final int progressBarNew = line / (lineCount / 10) * 10;
				if (progressBar != progressBarNew) {
					progressBar = progressBarNew;
					// System.out.print("\r");
					System.out.print(String.format("\r\t... processing signals: %d%% ...", progressBar));
				}
				
				// to this row, add cells to all signals (if value exists)
				for (int i = 1; i < values.size(); i++) {
					final String stringValue = values.get(i);
					if (stringValue != null && !stringValue.isEmpty()) {
						
						final String signalId = header[i];
						// we know: the 0th column is the timestamp
						final double doubleValue = Double.parseDouble(stringValue);
						floatSignals.get(signalId).getDataPoints().add(new DtoFloatDataPoint(Long.valueOf(values.get(0)), doubleValue));
					}
					
				}
				
				line++;
			}
		}
		
		System.out.print(String.format("\r\t... processing signals: %d%% ...", 100));
		
		// finally create data context
		System.out.println("and now adding signals to data context:");
		final DependencyAnalysisData analysisData = DependencyAnalysisData.create();
		
		for (Map.Entry<String, FloatSignal> signal : floatSignals.entrySet()) {
			analysisData.addFloatSignal(signal.getKey(), ((FloatSignal) signal.getValue()).getDataPoints());
			System.out.println(String.format("\t... float signal '%s' has %s values", signal.getKey(), ((FloatSignal) signal.getValue()).getDataPoints().size()));
		}
		System.out.println("\t... data context is finished!");
		
		return analysisData;
	}
	
	public static <T> T readJson(final File file, final Class<T> clazz) throws JsonParseException, JsonMappingException, IOException {
		System.out.println(String.format("\t... reading file '%s' ...", file.getPath()));
		return MAPPER.readValue(file, clazz);
	}
	
	public static <T> void writeJson(final File file, final T obj) throws JsonGenerationException, JsonMappingException, IOException {
		System.out.println(String.format("\t... writing file '%s' ...", file.getPath()));
		Optional.ofNullable(file.getParentFile()).ifPresent(f -> f.mkdirs());
		MAPPER.writeValue(file, obj);
	}
	
}
