# this python script performs the same calculation as example_da.ipynb


import json
import csv
import os
import argparse
import datetime

from aivis_engine_v2_da_sdk_python import (
    EngineLogger,
    DependencyAnalysisSetup,
    DependencyAnalysisData,
    DependencyAnalysis,
    DtoFloatDataPoint,
)

# ---------------------------------------------------------------------------------
# Data location

ANALYSIS_DATA_FILE = "train_da.csv"


# Input and output folder specified via args
def from_args():
    parser = argparse.ArgumentParser(
        description="Example project for dependency analysis"
    )
    # data is expected in input folder
    parser.add_argument(
        "--input", "-i", type=str, nargs=1, required=True, help="Path to input folder"
    )
    # output files will be written to output folder
    parser.add_argument(
        "--output", "-o", type=str, nargs=1, required=True, help="Path to output folder"
    )
    args = parser.parse_args()
    # create output folder if not existing
    if not os.path.exists(args.output[0]):
        os.mkdir(args.output[0])
    return {
        "input_folder": args.input[0],
        "output_folder": args.output[0],
    }


# ---------------------------------------------------------------------------------
# Logger and Threads


class Logger(EngineLogger):
    def log(self, level, thread, module, message):
        print(
            "%s [%s] [%s] %s - %s"
            % (datetime.datetime.now(), level, thread, module, message)
        )


def setup_logger_and_threads():
    print("Initializing DependencyAnalysisSetup: logging and threads ...")
    DependencyAnalysisSetup.register_logger(Logger())
    max_threads = 0  # number of threads will be chosen automatically with 0
    DependencyAnalysisSetup.init_thread_count(max_threads)
    print("\t... done!")
    return


# ---------------------------------------------------------------------------------
# Reading and Writing


def read_json(filename):
    print("\t... reading file '%s' ..." % filename)
    with open(filename, "r") as file:
        return json.loads(file.read())


def write_json(filename, obj):
    print("\t... writing file '%s' ..." % filename)
    with open(filename, "w") as file:
        file.write(json.dumps(obj, indent=2))
    return


# Main data reading function
# Float signals are added to an aivis data context via their id and a list of DtoFloatDataPoint.
# (Analogously for boolean and string signals.)
# The rest of the following function can be adapted if the data comes in a different format as in this example, e.g. in parquet files.
def read_csv_convert_to_aivis_format(filename):
    print("\t... reading file '%s' ..." % filename)

    # create empty data context for analysis data
    data_context = DependencyAnalysisData.create()

    # get line numbers to show nice progress bar
    line_count = sum(1 for _ in open(filename))

    # read file
    with open(filename, newline="\n") as csvfile:
        # in this example, signal ids are in the header, first column is timestamps
        reader = csv.reader(csvfile, delimiter=",", quotechar='"')
        headers = next(reader)
        signals = list(map(lambda h: {"signal_id": h, "data_points": []}, headers[1:]))

        # show nice progress bar
        progress_bar = -1
        line = 0

        for row in reader:
            # show nice progress bar
            progess_bar_new = round(line / (line_count / 10)) * 10
            if progress_bar != progess_bar_new:
                progress_bar = progess_bar_new
                print("\r\t... processing signals: %d%% ..." % progress_bar, end="")
            line += 1

            # get timestamp of this row, which is in the csv's first column
            timestamp = int(row[0])

            # add data points to all signal's list (if value exists)
            for i, column in enumerate(row[1:], start=0):
                if column:
                    # in this example, all signals are float signals
                    value = float(column)
                    signals[i]["data_points"].append(DtoFloatDataPoint(timestamp, value))
        print("... reading from file is done!")

        print("\t... adding signals to data context:")
        # for larger archives it would be better to read and then immediately add to data context in chunks
        # e.g. signal per signal
        for signal in signals:
            data_context.add_float_signal(signal["signal_id"], signal["data_points"])
            print(
                "\t   ... Float Signal "
                + signal["signal_id"]
                + " had "
                + str(len(signal["data_points"]))
                + " data points"
            )

        print("... data context is ready!")
        return data_context


# ---------------------------------------------------------------------------------
# MAIN


def main():
    folder_config = from_args()  # parsing args as folder_config
    setup_logger_and_threads()

    print("Loading analysis data ...")
    analysis_data = read_csv_convert_to_aivis_format(
        os.path.join(folder_config["input_folder"], ANALYSIS_DATA_FILE)
    )

    # print some signal to illustrate type DtoFloatDataPoint
    print("Example signal!")
    some_signal_id = "S_6"
    print("Signal id:", some_signal_id)
    some_signal_data_points = analysis_data.get_float_signal(some_signal_id)
    print("Signal length:", len(some_signal_data_points))
    print("Signal entry number 42:", some_signal_data_points[42])
    print(
        "Signal entry number 42 timestamp (unix/millisec):",
        some_signal_data_points[42].timestamp,
    )
    print("Signal entry number 42 value:", some_signal_data_points[42].value)

    # ---- ANALYSIS ----

    print("Creating analysis ...")
    # build analysis config - for more options see documentation
    analysis_config = json.dumps(
        {
            "sampling": {"maximalSampleCount": 10000},
            "clustering": {"minimalClusterSize": 1, "outputMultipleZoomLevels": True},
        }
    )

    # create analysis and train the model
    analysis = DependencyAnalysis.create(analysis_data, analysis_config)
    print("\t... done!")

    print("Writing output files...")
    analysis_report = json.loads(analysis.get_report())  # analysis infos
    write_json(
        os.path.join(folder_config["output_folder"], "analysis-report.json"),
        analysis_report,
    )
    print("\t... done!")

    print("Destroy everything ...")
    # destroying can be done once the respective object is not needed anymore - do this if memory is an issue
    analysis_data.destroy()
    analysis.destroy()

    print("\t... done!")

    # ---- Toolbox HTML Report ----
    from aivis_engine_v2_toolbox.api import build_dependency_analysis_report
    build_dependency_analysis_report(
        {"title": "Aivis Engine V2 - Dependency Analysis Example: Damadics",
         "analysisReportFile": os.path.join(folder_config["output_folder"], "analysis-report.json"),
         "outputFile": os.path.join(folder_config["output_folder"], "aivis-engine-v2-dependency-analysis-example.html")})

    print("ALL DONE!")

    return


if __name__ == "__main__":
    main()
